#include "box.h"
#include "frame.h"
#include "util.h"

Box::Box(Frame* p, const Vector& ctr, const Vector& s, double a,
		     const char* clr, const char* f, const char* t, 
		     unsigned short prop)
  : BaseBox(p, ctr, a, clr, f, t, prop)
{
  annuli = 1;
  size = new Vector[1];
  size[0] = s;

  strcpy(type,"box");
  handle = new Vector[4];
  numHandle = 4;

  updateBBox();
}

void Box::editBegin(int h)
{
  switch (h) {
  case 1:
    return;
  case 2:
    size[0] = Vector(-size[0][0],size[0][1]);
    return;
  case 3:
    size[0] = -size[0];
    return;
  case 4:
    size[0] = Vector(size[0][0],-size[0][1]);
    return;
  }
}

void Box::edit(const Vector& v, int h)
{
  // This sizes about the center
  //  size = ((v-center)*Rotate(angle)*2).abs();

  // This sizes about the opposite node
  Vector op = center+size[0]/2;
  size[0] = op - 
    (v * Translate(-center) * flip * Rotate(-angle) * Translate(center));
  center = v+((size[0]/2) * (Rotate(angle) * flip));

  updateBBox();
  doCallBacks(&editCB);
}

void Box::editEnd()
{
  size[0] = size[0].abs();

  updateBBox();
  doCallBacks(&editCB);
}

void Box::setSize(const Vector& s)
{
  size[0] = s;

  updateBBox();
  doCallBacks(&editCB);
}

// private

void Box::updateHandles()
{
  Vector s = size[0];
  if (properties & FIXED)
    s /= parent->getZoom();

  Matrix m = Rotate(angle) * flip * Translate(center) * parent->refToCanvas;
  handle[0] = Vector(-s[0],-s[1])/2 * m;
  handle[1] = Vector( s[0],-s[1])/2 * m;
  handle[2] = Vector( s[0], s[1])/2 * m;
  handle[3] = Vector(-s[0], s[1])/2 * m;
}

// list

void Box::list(ostream& str, int sys, int format, const char* delim)
{
  listCoordSystem(str, sys);
  str << ';';

  if (!(properties&INCLUDE))
    str << '-';

  switch ((Frame::CoordSystem)sys) {
  case Frame::IMAGE:
  case Frame::PHYSICAL:
  case Frame::DETECTOR:
  case Frame::AMPLIFIER:
    {
      Vector s = parent->mapLenFromRef(size[0], (Frame::CoordSystem)sys);
      Vector v = parent->map(center,Frame::REF,(Frame::CoordSystem)sys);
      str << type << '(' << setprecision(8) << v[0] << ',' << v[1] << ',' 
	  << s[0] << ',' << s[1] << ',' << radToDeg(angle) << ')';
    }
    break;

  case Frame::EQUATORIAL_FK4:
  case Frame::EQUATORIAL_FK5:
  case Frame::ECLIPTIC:
  case Frame::GALACTIC:
  case Frame::LINEAR:
  case Frame::ICRS:
    Vector s = parent->mapLenFromRef(size[0],
				  (Frame::CoordSystem)sys,
				  Frame::ARCSEC);
    switch ((Frame::CoordFormat)format) {
    case Frame::DEGREES:
      {
	Vector v = parent->map(center,Frame::REF,(Frame::CoordSystem)sys);
	str << type << '(' << setprecision(8) << v[0] << ',' << v[1] << ',' 
	    << s[0] << "\"" << ',' << s[1] << "\"" << ',' 
	    << radToDeg(angle) << ')';
      }
      break;
    case Frame::SEXAGESIMAL:
      {
	char buf[64];
	parent->mapToWCSStr(center, Frame::REF, (Frame::CoordSystem)sys,
			    (Frame::CoordFormat)format, buf, 64);
	char ra[16];
	char dec[16];
	istrstream wcs(buf,64);
	wcs >> ra >> dec;
	str << type << '(' << ra << ',' << dec << ',' 
	    << s[0] << "\"" << ',' << s[1] << "\"" << ',' 
	    << radToDeg(angle) << ')';
      }
      break;
    }
    break;
  }

  listProperties(str);
  str << delim;
}

void Box::listCiao(ostream& str, int sys, int format, const char* delim)
{
  if (!(properties&INCLUDE))
    str << '-';

  // radius is always in image coords

  switch ((Frame::CoordSystem)sys) {
  case Frame::IMAGE:
  case Frame::PHYSICAL:
  case Frame::DETECTOR:
  case Frame::AMPLIFIER:
    {
      Vector v = parent->map(center,Frame::REF,(Frame::CoordSystem)sys);
      Vector s = parent->mapLenFromRef(size[0], (Frame::CoordSystem)sys);
      str << "rotbox" << '(' << setprecision(8) << v[0] << ',' << v[1] << ',' 
	  << s[0] << ',' << s[1] << ',' << radToDeg(angle) << ')';
    }
    break;
  case Frame::EQUATORIAL_FK4:
  case Frame::EQUATORIAL_FK5:
  case Frame::ECLIPTIC:
  case Frame::GALACTIC:
  case Frame::LINEAR:
  case Frame::ICRS:
    Vector s = parent->mapLenFromRef(size[0],
				  (Frame::CoordSystem)sys,
				  Frame::ARCSEC);
    switch ((Frame::CoordFormat)format) {
    case Frame::DEGREES:
      {
	Vector v = parent->map(center,Frame::REF,(Frame::CoordSystem)sys);
	str << type << '(' << setprecision(8) << v[0] << "d," << v[1] << "d," 
	    << s[0] << "\"" << ',' << s[1] << "\"" << ',' 
	    << radToDeg(angle) << ')';
      }
      break;
    case Frame::SEXAGESIMAL:
      {
	char buf[64];
	parent->mapToWCSStr(center, Frame::REF, (Frame::CoordSystem)sys,
			    (Frame::CoordFormat)format, buf, 64);
	char ra[16];
	char dec[16];
	istrstream wcs(buf,64);
	wcs >> ra >> dec;
	str << type << '(' << ra << ',' << dec << ',' 
	    << s[0] << "\"" << ',' << s[1] << "\"" << ',' 
	    << radToDeg(angle) << ')';
      }
      break;
    }
    break;
  }

  str << delim;
}

void Box::listFtools(ostream& str, int sys, int format, const char* delim)
{
  if (!(properties&INCLUDE))
    str << '-';

  Vector v = parent->map(center,Frame::REF,(Frame::CoordSystem)sys);
  Vector s = parent->mapLenFromRef(size[0], (Frame::CoordSystem)sys);
  str << type << '(' << setprecision(8) << v[0] << ',' << v[1] << ',' 
      << s[0] << ',' << s[1] << ',' << radToDeg(angle) << ')';

  str << delim;
}

void Box::listPros(ostream& str, int sys, int format, const char* delim)
{
  switch ((Frame::CoordSystem)sys) {
  case Frame::IMAGE:
  case Frame::DETECTOR:
  case Frame::AMPLIFIER:
    {
      Vector s = parent->mapLenFromRef(size[0], (Frame::CoordSystem)sys);
      Vector v = parent->map(center,Frame::REF,Frame::IMAGE);
      str << "rotbox " << setprecision(8) << v << s << radToDeg(angle);
    }
    break;

  case Frame::PHYSICAL:
    {
      str << "physical; ";
      Vector s = parent->mapLenFromRef(size[0], (Frame::CoordSystem)sys);
      Vector v = parent->map(center,Frame::REF,(Frame::CoordSystem)sys);
      str << "rotbox " << setprecision(8) << v << s << radToDeg(angle);
    }
    break;

  case Frame::EQUATORIAL_FK4:
  case Frame::EQUATORIAL_FK5:
  case Frame::ECLIPTIC:
  case Frame::GALACTIC:
  case Frame::LINEAR:
  case Frame::ICRS:
    listProsCoordSystem(str,sys);
    str << "; ";

    // display wcs radius in ARCSEC
    Vector s = parent->mapLenFromRef(size[0], 
				  (Frame::CoordSystem)sys,
				  Frame::ARCSEC);

    switch ((Frame::CoordFormat)format) {
    case Frame::DEGREES:
      {
	Vector v = parent->map(center,Frame::REF,(Frame::CoordSystem)sys);
	str << "rotbox " << setprecision(8) << v[0] << "d " << v[1] << "d " 
	    << s[0] << "\" " << s[1] << "\" " << radToDeg(angle);
      }
      break;
    case Frame::SEXAGESIMAL:
      {
	char buf[64];
	parent->mapToWCSStr(center, Frame::REF, (Frame::CoordSystem)sys,
			    (Frame::CoordFormat)format, buf, 64);
	char ra[16];
	char dec[16];
	istrstream wcs(buf,64);
	wcs >> ra >> dec;
	if (dec[0]=='+')
	  str << "rotbox " << ra << ' ' << dec+1 << ' '
	      << s[0] << "\" " << s[1] << "\" " << radToDeg(angle);
	else
	  str << "rotbox " << ra << ' ' << dec << ' '
	      << s[0] << "\" " << s[1] << "\" " << radToDeg(angle);
      }
      break;
    }
    break;
  }

  str << delim;
}

void Box::listSAOtng(ostream& str, int sys, int format, const char* delim)
{
  if (text[0])
    str << '#' << text << delim;

  if (properties&INCLUDE)
    str << '+';
  else
    str << '-';

  // radius is always in image coords

  switch ((Frame::CoordSystem)sys) {
  case Frame::IMAGE:
  case Frame::DETECTOR:
  case Frame::AMPLIFIER:
    {
      Vector v = parent->map(center,Frame::REF,Frame::IMAGE);
      Vector s = parent->mapLenFromRef(size[0],Frame::IMAGE);
      str << type << '(' << setprecision(8) << v[0] << ',' << v[1] << ',' 
	  << s[0] << ',' << s[1] << ',' << radToDeg(angle) << ')';
    }
    break;
  case Frame::PHYSICAL:
    {
      Vector v = parent->map(center,Frame::REF,Frame::PHYSICAL);
      Vector s = parent->mapLenFromRef(size[0],Frame::PHYSICAL);
      str << type << '(' << setprecision(8) << v[0] << ',' << v[1] << ',' 
	  << s[0] << ',' << s[1] << ',' << radToDeg(angle) << ')';
    }
    break;
  case Frame::EQUATORIAL_FK4:
  case Frame::EQUATORIAL_FK5:
  case Frame::ECLIPTIC:
  case Frame::GALACTIC:
  case Frame::LINEAR:
  case Frame::ICRS:
    switch ((Frame::CoordFormat)format) {
    case Frame::DEGREES:
      {
	Vector v = parent->map(center,Frame::REF,(Frame::CoordSystem)sys);
	Vector s = parent->mapLenFromRef(size[0], Frame::IMAGE);
	str << type << '(' << setprecision(8) << v[0] << ',' << v[1] << ',' 
	    << s[0] << ',' << s[1] << ',' << radToDeg(angle) << ')';
      }
      break;
    case Frame::SEXAGESIMAL:
      {
	char buf[64];
	parent->mapToWCSStr(center, Frame::REF, (Frame::CoordSystem)sys,
			    (Frame::CoordFormat)format, buf, 64);
	Vector s = parent->mapLenFromRef(size[0], Frame::IMAGE);
	char ra[16];
	char dec[16];
	istrstream wcs(buf,64);
	wcs >> ra >> dec;
	str << type << '(' << ra << ',' << dec << ',' 
	    << s[0] << ',' << s[1] << ',' << radToDeg(angle) << ')';
      }
      break;
    }
    break;
  }

  str << " # " << colorName << delim;
}

void Box::listSAOimage(ostream& str, const char* delim)
{
  // all coords are in image coords

  if (!(properties&INCLUDE))
    str << '-';

  Vector v = parent->map(center,Frame::REF,Frame::IMAGE);
  str << type << '(' << setprecision(8) << v[0] << ',' << v[1] 
      << ',' << size[0][0] << ',' << size[0][1] << ')';

  str << delim;
}
