#include <Xlib.h>
#include <Xutil.h>

#include "frametrue.h"
#include "fitsimage.h"
#include "nan.h"
#include "ps.h"
#include "util.h"

// FrameTrueColor Member Functions

FrameTrueColor::FrameTrueColor(Tcl_Interp* i, Tk_Canvas c, Tk_Item* item)
  : Frame(i, c, item)
{
  skipUpdate = 0;

  colormapData = NULL;
  colormapPM = 0;
  colormapXM = NULL;
  colormapGCXOR = 0;
}

FrameTrueColor::~FrameTrueColor()
{
}

void FrameTrueColor::clearPixmap(Pixmap pmap, XImage* xmap)
{
  memset(xmap->data, 255, xmap->bytes_per_line * xmap->height);
  XPutImage(display, pmap, gc, xmap, 0, 0, 0, 0, 
	    options->width, options->height);
}

void FrameTrueColor::fillXImage(XImage* xmap, 
				Matrix& (FitsImage::*getMatrix)())
{
  // we need a colorScale before we can render

  if (!colorScale)
    return;

  // Xfer pixels from base image into map image--
  // This section of code has been optimized. First, the matix multiply
  // done inline, only using elements that count. (that is the first 6)
  // Second, the pixel xfer assumes 1 byte/pixel, and array indexing is
  // kepted to a minimum.

  int width = xmap->width;
  int height = xmap->height;

  int bytesPerPixel = xmap->bits_per_pixel/8;

  // clear ximage

  memset(xmap->data, 255, xmap->bytes_per_line * height);

  Matrix& m = (fits->*getMatrix)();
  register double m00 = m.matrix(0,0);
  register double m10 = m.matrix(1,0);
  register double m20 = m.matrix(2,0);
  register double m01 = m.matrix(0,1);
  register double m11 = m.matrix(1,1);
  register double m21 = m.matrix(2,1);
  register int srcWidth = fits->width();
  register int srcHeight = fits->height();

  register int length = colorScale->size() - 1;
  register int last = length * bytesPerPixel;
  register const unsigned char* table = colorScale->colors();

  register double ll = fits->getLowDouble();
  register double hh = fits->getHighDouble();
  register double diff = hh - ll;

  for (int j=0; j<height; j++) {
    // the line may be padded at the end
    register char* dest = xmap->data + j*xmap->bytes_per_line;

    for (int i=0; i<width; i++, dest+=bytesPerPixel) {
      // test as doubles, convert to int's at last minute.
      // use double to maintain precision
      // This solves a host of problems.

      register double srcX = i*m00 + j*m10 + m20;
      register double srcY = i*m01 + j*m11 + m21;

      if (srcX >= 0 && srcX < srcWidth && srcY >= 0 && srcY < srcHeight) {
	register double value = fits->getValueDouble(((int)srcY)*srcWidth + 
						     (int)srcX);
	
	if (!isnand(value)) {
	  if (value <= ll)
	    memcpy(dest, table, bytesPerPixel);
	  else if (value >= hh)
	    memcpy(dest, table+last, bytesPerPixel);
	  else
	    memcpy(dest, table+((int)(((value - ll)/diff * length) + .5))*
		   bytesPerPixel,bytesPerPixel);
	}
      }
    }
  }
}

void FrameTrueColor::fillXImageMosaic(XImage* xmap, 
				      Matrix& (FitsImage::*getMatrix)())
{
  // we need a colorScale before we can render

  if (!colorScale)
    return;

  // Xfer pixels from base image into map image--
  // This section of code has been optimized. First, the matix multiply
  // done inline, only using elements that count. (that is the first 6)
  // Second, the pixel xfer assumes 1 byte/pixel, and array indexing is
  // kepted to a minimum.

  // We cycle thru each pixel on the widget. For each pixel, 
  // we cycle thru each fits until we've run out of fits or 
  // we have a hit. If we hit, we remember it and use it until we
  // miss again. If we do not find a hit, we go on to the next pixel.

  int& width = xmap->width;
  int& height = xmap->height;

  int bytesPerPixel = xmap->bits_per_pixel/8;

  // clear ximage

  memset(xmap->data, 255, xmap->bytes_per_line * height);

  register int length = colorScale->size() - 1;
  register int last = length * bytesPerPixel;
  register const unsigned char* table = colorScale->colors();

  register double m00;
  register double m10;
  register double m20;
  register double m01;
  register double m11;
  register double m21;
  register int srcWidth;
  register int srcHeight;

  register double ll;
  register double hh;
  register double diff;

  // test as doubles, convert to int's at last minute.
  // use double to maintain precision
  // This solves a host of problems.

  int found = 0;
  FitsImage* current = fits;
  FitsImage* ptr;

  for (int j=0; j<height; j++) {
    // the line may be padded at the end
    register char* dest = xmap->data + j*xmap->bytes_per_line;

    for (int i=0; i<width; i++, dest+=bytesPerPixel) {
    tryagain32:
      if (!found) {
	ptr = current;
	current = current->next();

	Matrix& m = (ptr->*getMatrix)();
	m00 = m.matrix(0,0);
	m10 = m.matrix(1,0);
	m20 = m.matrix(2,0);
	m01 = m.matrix(0,1);
	m11 = m.matrix(1,1);
	m21 = m.matrix(2,1);
	srcWidth = ptr->width();
	srcHeight = ptr->height();

	ll = ptr->getLowDouble();
	hh = ptr->getHighDouble();
	diff = hh - ll;
      }

      register double srcX = i*m00 + j*m10 + m20;
      register double srcY = i*m01 + j*m11 + m21;

      if (srcX >= 0 && srcX < srcWidth && srcY >= 0 && srcY < srcHeight) {
	register double value = ptr->getValueDouble(((int)srcY)*srcWidth + 
						    (int)srcX);

	if (!isnand(value)) {
	  if (value <= ll)
	    memcpy(dest, table, bytesPerPixel);
	  else if (value >= hh)
	    memcpy(dest, table+last, bytesPerPixel);
	  else
	    memcpy(dest,
		   table+((int)(((value - ll)/diff * length) + .5))*
		   bytesPerPixel,bytesPerPixel);
	}
	found = 1;
      }
      else {
	found = 0;
	if (current)
	  goto tryagain32;
	else
	  current = fits;
      }
    }
  }
}

void FrameTrueColor::updateMagnifierReplicate(const Vector& v)
{
  // v is in CANVAS Coords
  // we have a race condition here. It's possible that a magnifier command has
  // been generated right after the pixmap has been invalidated and before
  // a new XImage has been generated. So, bail, and try again.

  if (!baseXImage || !magnifierXImage)
    return;

  char* image = baseXImage->data;
  int imageBytesPerPixel = baseXImage->bits_per_pixel/8;
  int& imageBytesPerLine = baseXImage->bytes_per_line;

  char* mag = magnifierXImage->data;
  int magBytesPerPixel = magnifierXImage->bits_per_pixel/8;
  int& magBytesPerLine = magnifierXImage->bytes_per_line;

  // map upper left of magnifier to image

  Matrix m = Translate(-(v * canvasToWidget)) * ximageToMagnifier;
  Matrix n = m.invert();

  Vector ll = Vector(0,0) * n;

  // and init image row and column counters

  int imageY = (int)ll[1];

  // loop thru each magnifier pixel

  char* row = new char[magBytesPerLine];
  for (int j=0; j<=magnifierHeight; j+=(int)magnifierZoom, imageY++) {

    // Clear tmp row

    memset(row,0,magBytesPerLine);

    // Build tmp row

    int imageX = (int)ll[0];
    for (int i=0; i<=magnifierWidth; i+=(int)magnifierZoom, imageX++) {

      // find pixel value

      unsigned char* value = new unsigned char[magBytesPerPixel];
      if (imageX>=0 && imageX<options->width && imageY>=0 
	  && imageY<options->height)
	memcpy(value, image + imageY*imageBytesPerLine + 
	       imageX*imageBytesPerPixel, magBytesPerPixel);
      else
	memset(value, 0, magBytesPerPixel);

      // replicate pixel

      int start = i - (int)(magnifierZoom/2.);
      if (start < 0)
	start = 0;
      int stop = i + (int)(magnifierZoom/2.);
      if (stop > magnifierWidth)
	stop = magnifierWidth;

      for (int k=start; k<stop; k++)
	memcpy(row+k*magBytesPerPixel, value, magBytesPerPixel);

      delete value;
    }

    // and replicate temp row
    
    int start = j - (int)(magnifierZoom/2.);
    if (start < 0)
      start = 0;
    int stop = j + (int)(magnifierZoom/2.);
    if (stop > magnifierHeight)
      stop = magnifierHeight;
    for (int k=start; k<stop; k++)
	memcpy(mag+(k*magBytesPerLine), row, magBytesPerLine);
  }
  delete row;

  XPutImage(display, magnifierPixmap, gc, magnifierXImage, 0, 0, 0, 0, 
	    magnifierWidth, magnifierHeight);
}

void FrameTrueColor::colormapCmd(int id, float b, float c, int i, 
				 unsigned short* index, unsigned char* cells, 
				 int cnt)
{
  // first check for change

  if (cmapID == id && bias == b && contrast == c && invert == i && colorCells)
    return;

  // we got a change

  cmapID = id;
  bias = b;
  contrast = c;
  invert = i;

  updateColorCells(index, cells, cnt);
  updateColorScale();

  if (fitsCount) {
    update(BASE); // always update
    skipUpdate = 0;
  }
  else {
    if (!skipUpdate) {
      update(BASE); // update once
      skipUpdate = 1;
    }      
  }
}

void FrameTrueColor::colormapBeginCmd()
{
  colormapBeginCmd(Vector(options->width,options->height)/2*widgetToCanvas);
}

void FrameTrueColor::colormapBeginCmd(const Vector& w)
{
  // we need some fits data
  // we assume the colorScale length will not change during motion calls

  if (!fits)
    return;

  int width = options->width;
  int height = options->height;
  cmapOffset = Vector();
  if (cmapArea) {
    width = height = cmapArea;

    switch (cmapMode) {
    case CENTER:
      cmapOffset = (Vector(options->width,options->height) - 
		    Vector(width,height))/2;
      break;
    case CLICK: 
      {
	Vector where = w * canvasToWidget;
	if (where[0] >= 0 && where[0] < options->width &&
	    where[1] >= 0 && where[1] < options->height)
	  cmapOffset = where - Vector(width,height)/2;
	else
	  cmapOffset = (Vector(options->width,options->height) - 
			Vector(width,height))/2;
      }
      break;
    }

    if (cmapOffset[0] < 0) {
      width += (int)cmapOffset[0];
      cmapOffset[0] = 0;
    }
    if (cmapOffset[0]+width >= options->width)
      width = options->width - (int)cmapOffset[0];

    if (cmapOffset[1] < 0) {
      height += (int)cmapOffset[1];
      cmapOffset[1] = 0;
    }

    if (cmapOffset[1]+height >= options->height)
      height = options->height - (int)cmapOffset[1];
  }

  // Create XImage

  if (!(colormapXM = XGetImage(display, pixmap, 0, 0, 
			       width, height, AllPlanes, ZPixmap))) {
    cerr << "FrameTrueColor Internal Error: Unable to Create Colormap XImage" 
	 << endl;
    exit(1);
  }

  // Create Pixmap

  colormapPM = Tk_GetPixmap(display, Tk_WindowId(tkwin), width, height, depth);
  if (!colormapPM) {
    cerr << "FrameTrueColor Internal Error: Unable to Create Colormap Pixmap" 
	 << endl;
    exit(1);
  }

  // colormapGCXOR

  colormapGCXOR = XCreateGC(display, Tk_WindowId(tkwin), 0, NULL);

  // Create table index array

  if (colormapData)
    delete [] colormapData;
  colormapData = new long[width*height];

  if (!colormapData) {
    cerr << "FrameTrueColor Internal Error: Unable to alloc tmp data array" 
	 << endl;
    exit(1);
  }

  // fill data array

  switch (fitsCount) {
  case 1:
    colormapBeginSingle(width, height);
    break;
  default:;
    colormapBeginMosaic(width, height);
    break;
  }
}

void FrameTrueColor::colormapBeginSingle(int width, int height)
{
  int xoff = (int)cmapOffset[0];
  int yoff = (int)cmapOffset[1];

  int bytesPerPixel = colormapXM->bits_per_pixel/8;

  Matrix& m = fits->getWidgetToData();
  register double m00 = m.matrix(0,0);
  register double m10 = m.matrix(1,0);
  register double m20 = m.matrix(2,0);
  register double m01 = m.matrix(0,1);
  register double m11 = m.matrix(1,1);
  register double m21 = m.matrix(2,1);
  register int srcWidth = fits->width();
  register int srcHeight = fits->height();

  register int length = colorScale->size() - 1;
  register int last = length * bytesPerPixel;

  register double ll = fits->getLowDouble();
  register double hh = fits->getHighDouble();
  register double diff = hh - ll;

  long* dest = colormapData;
  for (int j=yoff; j<yoff+height; j++) {
    for (int i=xoff; i<xoff+width; i++, dest++) {
      register double srcX = i*m00 + j*m10 + m20;
      register double srcY = i*m01 + j*m11 + m21;

      if (srcX >= 0 && srcX < srcWidth && srcY >= 0 && srcY < srcHeight) {
	register double value = fits->getValueDouble(((int)srcY)*srcWidth + 
						     (int)srcX);
	
	if (!isnand(value)) {
	  if (value <= ll)
	    *dest = 0;
	  else if (value >= hh)
	    *dest = last;
	  else
	    *dest = (int)(((value - ll)/diff * length) + .5)*bytesPerPixel;
	}
	else
	  *dest = -1;
      }
      else
	*dest = -1;
    }
  }
}

void FrameTrueColor::colormapBeginMosaic(int width, int height)
{
  int xoff = (int)cmapOffset[0];
  int yoff = (int)cmapOffset[1];

  int bytesPerPixel = colormapXM->bits_per_pixel/8;

  register int length = colorScale->size() - 1;
  register int last = length * bytesPerPixel;

  register double m00;
  register double m10;
  register double m20;
  register double m01;
  register double m11;
  register double m21;
  register int srcWidth;
  register int srcHeight;

  register double ll;
  register double hh;
  register double diff;

  int found = 0;
  FitsImage* current = fits;
  FitsImage* ptr;

  // scan fits

  long* dest = colormapData;
  for (int j=yoff; j<yoff+height; j++) {
    for (int i=xoff; i<xoff+width; i++, dest++) {
      *dest = -1; // initialize

    tryagainMos:
      if (!found) {
	ptr = current;
	current = current->next();

	Matrix& m = ptr->getWidgetToData();
	m00 = m.matrix(0,0);
	m10 = m.matrix(1,0);
	m20 = m.matrix(2,0);
	m01 = m.matrix(0,1);
	m11 = m.matrix(1,1);
	m21 = m.matrix(2,1);
	srcWidth = ptr->width();
	srcHeight = ptr->height();

	ll = ptr->getLowDouble();
	hh = ptr->getHighDouble();
	diff = hh - ll;
      }

      register double srcX = i*m00 + j*m10 + m20;
      register double srcY = i*m01 + j*m11 + m21;

      if (srcX >= 0 && srcX < srcWidth && srcY >= 0 && srcY < srcHeight) {
	register double value = ptr->getValueDouble(((int)srcY)*srcWidth + 
						    (int)srcX);

	if (!isnand(value)) {
	  if (value <= ll)
	    *dest = 0;
	  else if (value >= hh)
	    *dest = last;
	  else
	    *dest = (int)(((value - ll)/diff * length) + .5)*bytesPerPixel;
	}
	found = 1;
      }
      else {
	found = 0;
	if (current)
	  goto tryagainMos;
	else
	  current = fits;
      }
    }
  }
}

void FrameTrueColor::colormapMotionCmd(int id, float b, float c, int i, 
				       unsigned short* index, 
				       unsigned char* cells, int cnt)
{
  // first check for change

  if (cmapID == id && bias == b && contrast == c && invert == i && colorCells)
    return;

  // we got a change

  cmapID = id;
  bias = b;
  contrast = c;
  invert = i;

  updateColorCells(index, cells, cnt);
  updateColorScale();

  // if we have no data, stop now

  if (!fits) 
    return;

  // clear ximage

  int& width = colormapXM->width;
  int& height = colormapXM->height;

  memset(colormapXM->data, 255, colormapXM->bytes_per_line * height);

  int bytesPerPixel = colormapXM->bits_per_pixel/8;
  int& bytesPerLine = colormapXM->bytes_per_line;

  register const unsigned char* table = colorScale->colors();

  register long* src = colormapData;
  for (int j=0; j<height; j++) {
    // line may be padded at end
    register char* dest = colormapXM->data + j*bytesPerLine;

    for (int i=0; i<width; i++, src++, dest+=bytesPerPixel)
      if (*src >=0)
	memcpy(dest, table+(*src), bytesPerPixel);
  }

  // XImage to Pixmap

  XPutImage(display, colormapPM, gc, colormapXM, 0, 0, 0, 0, width, height);

  // Display Pixmap

  Vector d = cmapOffset * widgetToWindow;
  XCopyArea(display, colormapPM, Tk_WindowId(tkwin), colormapGCXOR, 0, 0, 
	    width, height, (int)d[0], (int)d[1]);
}

void FrameTrueColor::colormapEndCmd(int id, float b, float c, int i, 
				    unsigned short* index, 
				    unsigned char* cells, int cnt)
{
  if (colormapXM) {
    XDestroyImage(colormapXM);
    colormapXM = NULL;
  }

  if (colormapPM) {
    Tk_FreePixmap(display, colormapPM);
    colormapPM = 0;
  }

  if (colormapData) {
    delete [] colormapData;
    colormapData = NULL;
  }

  if (colormapGCXOR) {
    XFreeGC(display, colormapGCXOR);
    colormapGCXOR = 0;
  }

  if (fits)
    update(BASE); // always update
}

void FrameTrueColor::rotateMotion()
{
  // Rotate from src to dest

  Vector c = Vector(options->width,options->height)/2;
  Matrix m = Translate(-c) * Rotate(rotation-rotateRotation) * Translate(c);
  Matrix mm = m.invert();
  register int width = options->width;
  register int height = options->height;
  register double m00 = mm.matrix(0,0);
  register double m10 = mm.matrix(1,0);
  register double m20 = mm.matrix(2,0);
  register double m01 = mm.matrix(0,1);
  register double m11 = mm.matrix(1,1);
  register double m21 = mm.matrix(2,1);
  register char* src = rotateSrcXM->data;

  int bytesPerPixel = rotateDestXM->bits_per_pixel/8;

  for (register int j=0; j<height; j++) {
    // the line may be padded at the end
    register char* dest = rotateDestXM->data + j*rotateDestXM->bytes_per_line;

    for (register int i=0; i<width; i++, dest+=bytesPerPixel) {
      register double srcX = i*m00 + j*m10 + m20;
      register double srcY = i*m01 + j*m11 + m21;

      if (srcX >= 0 && srcX < width && srcY >= 0 && srcY < height)
	memcpy(dest, 
	       src + ((int)srcY)*rotateDestXM->bytes_per_line+
	       ((int)srcX)*bytesPerPixel, bytesPerPixel);
      else
	memset(dest, 255, bytesPerPixel);
    }
  }

  // XImage to Pixmap

  XPutImage(display, rotatePM, gc, rotateDestXM,
	    0, 0, 0, 0, options->width, options->height);

  // Display Pixmap

  Vector d = Vector(0,0) * widgetToWindow;
  XCopyArea(display, rotatePM, Tk_WindowId(tkwin), rotateGCXOR, 0, 0, 
	    options->width, options->height, (int)d[0], (int)d[1]);
}
