#include <stdlib.h>
#include <string.h>

#include <iostream.h>
#include <strstream.h>

#include "frame.h"
#include "fitsimage.h"
#include "util.h"
#include "mmap.h"

#include "marker.h"
#include "basemarker.h"
#include "basecircle.h"
#include "annulus.h"
#include "circle.h"
#include "ellipse.h"
#include "ellipseannulus.h"
#include "box.h"
#include "boxannulus.h"
#include "polygon.h"
#include "line.h"
#include "text.h"
#include "ruler.h"
#include "point.h"

// NOTE: all marker traversal routines use a local ptr as opposed to the
// list current() because marker call backs may invoke another traversal 
// routine or the event loop, which may process pending events

// Marker Parser Stuff

#undef yyFlexLexer
#define yyFlexLexer mkFlexLexer
#include <FlexLexer.h>

mkFlexLexer* mklexx = NULL;    // used by mklex
extern Frame* fr;              // used by mkerror
extern int mkparse(void*);
int mkmode = Frame::DS9;       // used by mkparser

int mklex()
{
  return (mklexx ? mklexx->yylex() : 0);
}

void mkerror(const char* m)
{
  if (fr) {
    fr->error(m);
    const char* cmd = mklexx ? mklexx->YYText() : (const char*)NULL;
    if (cmd && cmd[0] != '\n') {
      fr->error(": ");
      fr->error(cmd);
    }
  }
}

#undef yyFlexLexer
#define yyFlexLexer othFlexLexer
#include <FlexLexer.h>

othFlexLexer* othlexx = NULL;    // used by othlex
extern Frame* fr;              // used by otherror
extern int othparse(void*);

int othlex()
{
  return (othlexx ? othlexx->yylex() : 0);
}

void otherror(const char* m)
{
  if (fr) {
    fr->error(m);
    const char* cmd = othlexx ? othlexx->YYText() : (const char*)NULL;
    if (cmd && cmd[0] != '\n') {
      fr->error(": ");
      fr->error(cmd);
    }
  }
}

// Marker Commands

void Frame::createCircleCmd(const Vector& center, double radius,
			    const char* color, const char* font,
			    const char* text, unsigned short prop)
{
  Marker* m = new Circle(this, center, radius, color, font, text, prop);
  createMarker(m);
}

void Frame::createCircleCmd(const Vector& center, CoordSystem sys,
			    double radius, CoordFormat format,
			    const char* color, const char* font,
			    const char* text, unsigned short prop)
{
  double r = mapLenToRef(radius,sys,format);
  Marker* m = new Circle(this, map(center,sys,REF), r, 
			 color, font, text, prop);
  createMarker(m);
}

void Frame::createAnnulusCmd(const Vector& center, double radius,
			     const char* color, const char* font,
			     const char* text, unsigned short prop)
{
  Marker* m = new Annulus(this, center, radius, color, font, text, prop);
  createMarker(m);
}

void Frame::createAnnulusCmd(const Vector& center, CoordSystem sys,
			     double radius, CoordFormat format,
			     const char* color, const char* font,
			     const char* text, unsigned short prop)
{
  double r = mapLenToRef(radius,sys,format);
  Marker* m = new Annulus(this, map(center,sys,REF), r, 
			  color, font, text, prop);
  createMarker(m);
}

void Frame::createAnnulusCmd(const Vector& center, CoordSystem sys,
			     double start, double stop, int num,
			     CoordFormat format,
			     const char* color, const char* font,
			     const char* text, unsigned short prop)
{
  double r1 = mapLenToRef(start,sys,format);
  double r2 = mapLenToRef(stop,sys,format);
  Marker* m = new Annulus(this, map(center,sys,REF), r1, r2, num,
			  color, font, text, prop);
  createMarker(m);
}

void Frame::createAnnulusCmd(const Vector& center,
			     double start, double stop, int num,
			     const char* color, const char* font,
			     const char* text, unsigned short prop)
{
  Marker* m = new Annulus(this, center, start, stop, num,
			  color, font, text, prop);
  createMarker(m);
}

void Frame::createAnnulusCmd(const Vector& center,
			     double* radii, int num,
			     const char* color, const char* font,
			     const char* text, unsigned short prop)
{
  Marker* m = new Annulus(this, center, radii, num,
			  color, font, text, prop);
  createMarker(m);
}

void Frame::createAnnulusCmd(const Vector& center, CoordSystem sys,
			     double* radii, int num,
			     CoordFormat format,
			     const char* color, const char* font,
			     const char* text, unsigned short prop)
{
  for (int i=0; i<num; i++)
    radii[i] = mapLenToRef(radii[i],sys,format);

  Marker* m = new Annulus(this, map(center,sys,REF), radii, num,
			  color, font, text, prop);
  createMarker(m);
}

void Frame::createBoxCmd(const Vector& center, const Vector& size, 
			 double angle,
			 const char* color, const char* font,
			 const char* text, unsigned short prop)
{
  Marker* m = new Box(this, center, size, angle, color, font, text, prop);
  createMarker(m);
}

void Frame::createBoxCmd(const Vector& center, CoordSystem sys,
			 const Vector& size, CoordFormat format,
			 double angle,
			 const char* color, const char* font,
			 const char* text, unsigned short prop)
{
  Marker* m = new Box(this, map(center,sys,REF), 
		      mapLenToRef(size,sys,format),
		      angle, color, font, text, prop);
  createMarker(m);
}

void Frame::createBoxAnnulusCmd(const Vector& center,
				const Vector& inner, const Vector& outer,
				int num,
				double angle,
				const char* color, const char* font, 
				const char* text, unsigned short prop)
{
  Marker* m = new BoxAnnulus(this, center, inner, outer, num, angle,
			     color, font, text, prop);
  createMarker(m);
}

void Frame::createBoxAnnulusCmd(const Vector& center, CoordSystem sys,
				const Vector& inner, const Vector& outer,
				int num,
				CoordFormat format, double angle,
				const char* color, const char* font, 
				const char* text, unsigned short prop)
{
  Marker* m = new BoxAnnulus(this, map(center,sys,REF), 
			     mapLenToRef(inner,sys,format),
			     mapLenToRef(outer,sys,format),
			     num, angle,
			     color, font, text, prop);
  createMarker(m);
}

void Frame::createBoxAnnulusCmd(const Vector& center, CoordSystem sys,
				double angle, int num,
				CoordFormat format, Vector* size,
				const char* color, const char* font, 
				const char* text, unsigned short prop)
{
  for (int i=0; i<num; i++)
    size[i] = mapLenToRef(size[i],sys,format);

  Marker* m = new BoxAnnulus(this, map(center,sys,REF), angle, num, size,
			     color, font, text, prop);
  createMarker(m);
}

void Frame::createBoxAnnulusCmd(const Vector& center, double angle, int num,
				Vector* size,
				const char* color, const char* font, 
				const char* text, unsigned short prop)
{
  Marker* m = new BoxAnnulus(this, center, angle, num, size,
			     color, font, text, prop);
  createMarker(m);
}

void Frame::createEllipseCmd(const Vector& center, const Vector& radius,
			     double angle,
			     const char* color, const char* font,
			     const char* text, unsigned short prop)
{
  Marker* m = new Ellipse(this, center, radius, angle,
			  color, font, text, prop);
  createMarker(m);
}

void Frame::createEllipseCmd(const Vector& center, CoordSystem sys,
			     const Vector& radius, CoordFormat format,
			     double angle,
			     const char* color, const char* font,
			     const char* text, unsigned short prop)
{
  Marker* m = new Ellipse(this, map(center,sys,REF), 
			  mapLenToRef(radius,sys,format),
			  angle, color, font, text, prop);
  createMarker(m);
}

void Frame::createEllipseAnnulusCmd(const Vector& center, CoordSystem sys,
				    const Vector& inner, const Vector& outer,
				    int num,
				    CoordFormat format, double angle,
				    const char* color, const char* font, 
				    const char* text, unsigned short prop)
{
  Marker* m = new EllipseAnnulus(this, map(center,sys,REF), 
				 mapLenToRef(inner,sys,format),
				 mapLenToRef(outer,sys,format),
				 num, angle,
				 color, font, text, prop);
  createMarker(m);
}

void Frame::createEllipseAnnulusCmd(const Vector& center,
				    const Vector& inner, const Vector& outer,
				    int num, double angle,
				    const char* color, const char* font, 
				    const char* text, unsigned short prop)
{
  Marker* m = new EllipseAnnulus(this, center, inner, outer, num, angle,
				 color, font, text, prop);
  createMarker(m);
}

void Frame::createEllipseAnnulusCmd(const Vector& center, CoordSystem sys,
				    double angle, int num,
				    CoordFormat format, Vector* radii,
				    const char* color, const char* font, 
				    const char* text, unsigned short prop)
{
  for (int i=0; i<num; i++)
    radii[i] = mapLenToRef(radii[i],sys,format);

  Marker* m = new EllipseAnnulus(this, map(center,sys,REF), angle, num, radii,
				 color, font, text, prop);
  createMarker(m);
}

void Frame::createEllipseAnnulusCmd(const Vector& center, double angle,
				    int num, Vector* radii,
				    const char* color, const char* font, 
				    const char* text, unsigned short prop)
{
  Marker* m = new EllipseAnnulus(this, center, angle, num, radii,
				 color, font, text, prop);
  createMarker(m);
}

void Frame::createPolygonCmd(const Vector& center, CoordSystem sys,
			     const Vector& bb, 			     
			     const char* color, const char* font,
			     const char* text, unsigned short prop)
{
  Marker* m = new Polygon(this, map(center,sys,REF), bb,
			  color, font, text, prop);
  createMarker(m);
}

void Frame::createPolygonCmd(const Vector& center, const Vector& bb,
			     const char* color, const char* font,
			     const char* text, unsigned short prop)
{
  Marker* m = new Polygon(this, center, bb, color, font, text, prop);
  createMarker(m);
}

void Frame::createPolygonCmd(List<Vertex>* list,
			     const char* color, const char* font,
			     const char* text, unsigned short prop)
{
  Marker* m = new Polygon(this, list, color, font, text, prop);
  createMarker(m);
}

void Frame::createPolygonCmd(List<Vertex>* list, CoordSystem sys,
			     const char* color, const char* font,
			     const char* text, unsigned short prop)
{
  list->head();
  do
    list->current()->vector = map(list->current()->vector, sys, REF);
  while (list->next());

  Marker* m = new Polygon(this, list, color, font, text, prop);
  createMarker(m);
}

void Frame::createLineCmd(const Vector& center, const Vector& p2, 
			  int arrow1, int arrow2,
			  const char* color, const char* font,
			  const char* text, unsigned short prop)
{
  Marker* m = new Line(this, center, p2, arrow1, arrow2,
		       color, font, text, prop);
  createMarker(m);
}

void Frame::createLineCmd(const Vector& center, const Vector& p2, 
			  CoordSystem sys,
			  int arrow1, int arrow2,
			  const char* color, const char* font,
			  const char* text, unsigned short prop)
{
  Marker* m = new Line(this, map(center,sys,REF), map(p2,sys,REF), 
		       arrow1, arrow2,
		       color, font, text, prop);
  createMarker(m);
}

void Frame::createTextCmd(const Vector& center,
			  const char* color, const char* font,
			  const char* text, unsigned short prop)
{
  Marker* m = new Text(this, center, color, font, text, prop);
  createMarker(m);
}

void Frame::createTextCmd(const Vector& center, CoordSystem sys,
			  const char* color, const char* font,
			  const char* text, unsigned short prop)
{
  Marker* m = new Text(this, map(center,sys,REF), color, font, text, prop);
  createMarker(m);
}

void Frame::createRulerCmd(const Vector& center, const Vector& p2,
			   CoordFormat format,
			   const char* color, const char* font,
			   const char* text, unsigned short prop)
{
  Marker* m = new Ruler(this, center, p2, format, color, font, text, prop);
  createMarker(m);
}

void Frame::createRulerCmd(const Vector& center, const Vector& p2,
			   CoordSystem sys, CoordFormat format,
			   const char* color, const char* font,
			   const char* text, unsigned short prop)
{
  Marker* m = new Ruler(this, map(center,sys,REF), map(p2,sys,REF), format,
			color, font, text, prop);
  createMarker(m);
}

void Frame::createCirclePointCmd(const Vector& center,
				 const char* color, const char* font,
				 const char* text, unsigned short prop)
{
  Marker* m = new CirclePoint(this, center, color, font, text, prop);
  createMarker(m);
}

void Frame::createCirclePointCmd(const Vector& center, CoordSystem sys,
				 const char* color, const char* font,
				 const char* text, unsigned short prop)
{
  Marker* m = new CirclePoint(this, map(center,sys,REF), 
			      color, font, text, prop);
  createMarker(m);
}

void Frame::createBoxPointCmd(const Vector& center,
			      const char* color, const char* font,
			      const char* text, unsigned short prop)
{
  Marker* m = new BoxPoint(this, center, color, font, text, prop);
  createMarker(m);
}

void Frame::createBoxPointCmd(const Vector& center, CoordSystem sys,
			      const char* color, const char* font,
			      const char* text, unsigned short prop)
{
  Marker* m = new BoxPoint(this, map(center,sys,REF), color, font, text, prop);
  createMarker(m);
}

void Frame::createDiamondPointCmd(const Vector& center,
				  const char* color, const char* font,
				  const char* text, unsigned short prop)
{
  Marker* m = new DiamondPoint(this, center, color, font, text, prop);
  createMarker(m);
}

void Frame::createDiamondPointCmd(const Vector& center, CoordSystem sys,
				  const char* color, const char* font,
				  const char* text, unsigned short prop)
{
  Marker* m = new DiamondPoint(this, map(center,sys,REF), 
			       color, font, text, prop);
  createMarker(m);
}

void Frame::createCrossPointCmd(const Vector& center,
				const char* color, const char* font,
				const char* text, unsigned short prop)
{
  Marker* m = new CrossPoint(this, center, color, font, text, prop);
  createMarker(m);
}

void Frame::createCrossPointCmd(const Vector& center, CoordSystem sys,
				const char* color, const char* font,
				const char* text, unsigned short prop)
{
  Marker* m = new CrossPoint(this, map(center,sys,REF), 
			     color, font, text, prop);
  createMarker(m);
}

void Frame::createExPointCmd(const Vector& center,
			     const char* color, const char* font,
			     const char* text, unsigned short prop)
{
  Marker* m = new ExPoint(this, center, color, font, text, prop);
  createMarker(m);
}

void Frame::createExPointCmd(const Vector& center, CoordSystem sys,
			     const char* color, const char* font,
			     const char* text, unsigned short prop)
{
  Marker* m = new ExPoint(this, map(center,sys,REF), color, font, text, prop);
  createMarker(m);
}

void Frame::createArrowPointCmd(const Vector& center,
				const char* color, const char* font,
				const char* text, unsigned short prop)
{
  Marker* m = new ArrowPoint(this, center, color, font, text, prop);
  createMarker(m);
}

void Frame::createArrowPointCmd(const Vector& center, CoordSystem sys,
				const char* color, const char* font,
				const char* text, unsigned short prop)
{
  Marker* m = new ArrowPoint(this, map(center,sys,REF), 
			     color, font, text, prop);
  createMarker(m);
}

void Frame::createBoxCirclePointCmd(const Vector& center,
				    const char* color, const char* font,
				    const char* text, unsigned short prop)
{
  Marker* m = new BoxCirclePoint(this, center, color, font, text, prop);
  createMarker(m);
}

void Frame::createBoxCirclePointCmd(const Vector& center, CoordSystem sys,
				    const char* color, const char* font,
				    const char* text, unsigned short prop)
{
  Marker* m = new BoxCirclePoint(this, map(center,sys,REF), 
				 color, font, text, prop);
  createMarker(m);
}

void Frame::createMarker(Marker* m)
{
  markers.append(m);

  // now update new marker

  update(PIXMAP, m->getAllBBox());

  // and return id

  printIntegerTcl(m->getId());
}

void Frame::getMarkerAnnulusRadiusCmd(int id, CoordFormat format)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      const double* r = ((Annulus*)m)->getRadii();
      int cnt = ((Annulus*)m)->getAnnuliCount();
      for (int i=0; i<cnt; i++) {
	Vector v = mapLenFromRef(Vector(r[i],0), format);
	printDoubleTcl(v[0], DEFAULT);
	Tcl_AppendResult(interp, "\n", NULL);
      }
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "", NULL);
}

void Frame::getMarkerBoxAnnulusSizeCmd(int id, CoordFormat format)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      const Vector* r = ((BoxAnnulus*)m)->getSize();
      int cnt = ((BoxAnnulus*)m)->getAnnuliCount();
      for (int i=0; i<cnt; i++) {
	Vector v = mapLenFromRef(r[i], format);
	printVectorTcl(v, DEFAULT);
	Tcl_AppendResult(interp, "\n", NULL);
      }
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "", NULL);
}

void Frame::getMarkerCenterCmd(int id, CoordSystem sys)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      printCoordTcl(m->getCenter(), REF, sys, DEFAULT);
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "", NULL);
}

void Frame::getMarkerCenterCmd(int id, CoordSystem wcs, CoordFormat format)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      getWCSCoordCmd(m->getCenter(), REF, wcs, format);
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "", NULL);
}

void Frame::getMarkerCircleRadiusCmd(int id, CoordFormat format)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      double r = ((Circle*)m)->getRadius();
      printDoubleTcl(mapLenFromRef(r, format), DEFAULT);
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "", NULL);
}

void Frame::getMarkerColorCmd()
{
  // return first found

  Marker* m=markers.head();
  while (m) {
    if (m->isSelected()) {
      Tcl_AppendResult(interp,  m->getColorName(), NULL);
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "", NULL);
}

void Frame::getMarkerColorCmd(int id)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      Tcl_AppendResult(interp,  m->getColorName(), NULL);
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "", NULL);
}

void Frame::getMarkerEllipseRadiusCmd(int id, CoordFormat format)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      Vector r = ((Ellipse*)m)->getRadius();
      printVectorTcl(mapLenFromRef(r,format),DEFAULT);
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "", NULL);
}

void Frame::getMarkerEllipseAngleCmd(int id)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      printDoubleTcl(radToDeg(((BaseEllipse*)m)->getAngle()), DEFAULT);
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "", NULL);
}

void Frame::getMarkerEllipseAnnulusRadiusCmd(int id, CoordFormat format)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      const Vector* r = ((EllipseAnnulus*)m)->getRadii();
      int cnt = ((EllipseAnnulus*)m)->getAnnuliCount();
      for (int i=0; i<cnt; i++) {
	Vector v = mapLenFromRef(r[i], format);
	printVectorTcl(v, DEFAULT);
	Tcl_AppendResult(interp, "\n", NULL);
      }
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "", NULL);
}

void Frame::getMarkerFontCmd()
{
  Marker* m=markers.head();
  while (m) {
    if (m->isSelected()) {
      Tcl_AppendResult(interp, m->getFont(), NULL);
      return;
    }
    m=m->next();
  }
}

void Frame::getMarkerFontCmd(int id)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      Tcl_AppendResult(interp, m->getFont(), NULL);
      return;
    }
    m=m->next();
  }
}

void Frame::getMarkerHandleCmd(const Vector& v)
{
  Marker* m=markers.head();
  while (m) {
    if (m->isSelected()) {
      int h = m->onHandle(v);
      if (h) {
	char buf [16];
	ostrstream str(buf,16);
	str << m->getId() << ' ' << h << ends;
	Tcl_AppendResult(interp, buf, NULL);
	return;
      }
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "0 0", NULL);
}

void Frame::getMarkerIdCmd(const Vector& v)
{
  // v is in canvas coords

  Marker* m=markers.head();
  while (m) {
    if (m->isIn(v)) {
      printIntegerTcl(m->getId());
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "0", NULL);
}

void Frame::getMarkerIdAllCmd()
{
  char buf[8];
  Marker* m=markers.head();
  while (m) {
    ostrstream str(buf,8);
    str << m->getId() << ' ' << ends;
    Tcl_AppendResult(interp, buf, NULL);
    m=m->next();
  }
}

void Frame::getMarkerLineCmd(int id, CoordSystem sys)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      printCoordTcl(((Line*)m)->getP1(), REF, sys, DEFAULT);
      printCoordTcl(((Line*)m)->getP2(), REF, sys, DEFAULT);
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "", NULL);
}

void Frame::getMarkerLineCmd(int id, CoordSystem wcs, CoordFormat format)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      getWCSCoordCmd(((Line*)m)->getP1(), REF, wcs, format);
      Tcl_AppendResult(interp, " ", NULL);
      getWCSCoordCmd(((Line*)m)->getP2(), REF, wcs, format);
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "", NULL);
}

void Frame::getMarkerLineArrowCmd(int id)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      if (((Line*)m)->getP1Arrow())
	Tcl_AppendResult(interp, "1", NULL);
      else
	Tcl_AppendResult(interp, "0", NULL);

      if (((Line*)m)->getP2Arrow())
	Tcl_AppendResult(interp, " 1", NULL);
      else
	Tcl_AppendResult(interp, " 0", NULL);

      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "", NULL);
}

void Frame::getMarkerLineDistanceCmd(int id, CoordFormat format)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      printDoubleTcl(((Line*)m)->getDistance(format), DEFAULT);
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "", NULL);
}

void Frame::getMarkerPolygonAngleCmd(int id)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      printDoubleTcl(radToDeg(((Polygon*)m)->getAngle()), DEFAULT);
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "", NULL);
}

void Frame::getMarkerPolygonSegmentCmd(const Vector& v)
{
  Marker* m=markers.head();
  while (m) {
    if (m->isSelected() && m->isPolygon()) {
      int s = ((Polygon*)(m))->getSegment(v);
      if (s) {
	char buf [16];
	ostrstream str(buf,16);
	str << m->getId() << ' ' << s << ends;
	Tcl_AppendResult(interp, buf, NULL);
	return;
      }
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "0 0", NULL);
}


void Frame::getMarkerPropertyCmd(unsigned short which)
{
  // return first selected found

  Marker* m=markers.head();
  while (m) {
    if (m->isSelected()) {
      m->getProperty(which) ?
	Tcl_AppendResult(interp, "1", NULL): 
	Tcl_AppendResult(interp, "0", NULL);
      return;
    }
    m=m->next();
  }

  // else, return 'off'

  Tcl_AppendResult(interp, "0", NULL);
}

void Frame::getMarkerPropertyCmd(int id, unsigned short which)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      m->getProperty(which) ?
	Tcl_AppendResult(interp, "1", NULL):
	Tcl_AppendResult(interp, "0", NULL);
      return;
    }
    m=m->next();
  }

  // else, return 'off'

  Tcl_AppendResult(interp, "0", NULL);
}

void Frame::getMarkerBoxSizeCmd(int id, CoordFormat format)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      Vector s = ((Box*)m)->getSize();
      printVectorTcl(mapLenFromRef(s,format), DEFAULT);
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "", NULL);
}

void Frame::getMarkerBoxAngleCmd(int id)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      printDoubleTcl(radToDeg(((Box*)m)->getAngle()), DEFAULT);
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "", NULL);
}

void Frame::getMarkerRulerDistanceCmd(int id)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      printDoubleTcl(((Ruler*)m)->getDistance(), DEFAULT);
      printVectorTcl(((Ruler*)m)->getAxisDistance(), DEFAULT);
      printDoubleTcl(radToDeg(((Ruler*)m)->getAngle()), DEFAULT);
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "", NULL);
}

void Frame::getMarkerRulerPointCmd(int id, CoordSystem sys)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      printCoordTcl(((Ruler*)m)->getP1(), REF, sys, DEFAULT);
      printCoordTcl(((Ruler*)m)->getP2(), REF, sys, DEFAULT);
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "", NULL);
}

void Frame::getMarkerRulerPointCmd(int id, CoordSystem wcs, CoordFormat format)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      getWCSCoordCmd(((Ruler*)m)->getP1(), REF, wcs, format);
      Tcl_AppendResult(interp, " ", NULL);
      getWCSCoordCmd(((Ruler*)m)->getP2(), REF, wcs, format);
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "", NULL);
}

void Frame::getMarkerSelectedCmd()
{
  char buf[8];
  Marker* m=markers.head();
  while (m) {
    if (m->isSelected()) {
      ostrstream str(buf,8);
      str << m->getId() << ' ' << ends;
      Tcl_AppendResult(interp, buf, NULL);
    }
    m=m->next();
  }
}

void Frame::getMarkerSelectedCmd(const Vector& v)
{
  char buf[8];
  Marker* m=markers.head();
  while (m) {
    if (m->isIn(v) && m->isSelected()) {
      ostrstream str(buf,8);
      str << m->getId() << ' ' << ends;
      Tcl_AppendResult(interp, buf, NULL);
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "0", NULL);
}

void Frame::getMarkerSelectedNumberCmd()
{
  int count=0;
  Marker* m=markers.head();
  while (m) {
    if (m->isSelected())
      count++;
    m=m->next();
  }

  printIntegerTcl(count);
}

void Frame::getMarkerTextCmd(int id)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      Tcl_AppendResult(interp, m->getText(), NULL);
      return;
    }
    m=m->next();
  }
}

void Frame::getMarkerTypeCmd(int id)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      Tcl_AppendResult(interp, m->getType(), NULL);
      return;
    }
    m=m->next();
  }
}

void Frame::markerAnnulusCreateRadiusCmd(int id, const Vector& v)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      printIntegerTcl(((Annulus*)m)->addRadii(v*canvasToRef));
      update(PIXMAP, m->getBBox());
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void Frame::markerAnnulusDeleteRadiusCmd(int id, int h)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      if (m->canEdit()) {
	// it may shrink
	update(PIXMAP, m->getAllBBox());
	((Annulus*)m)->deleteRadii(h);
	update(PIXMAP, m->getAllBBox());
      }
      return;
    }
    m=m->next();
  }
}

void Frame::markerAnnulusRadiusCmd(int id, double inner, double outer, int num,
				   CoordFormat format)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      // it may shrink
      update(PIXMAP, m->getAllBBox());
      double r1 = mapLenToRef(inner,format);
      double r2 = mapLenToRef(outer,format);
      ((Annulus*)m)->setRadii(r1,r2,num);
      update(PIXMAP, m->getAllBBox());
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void Frame::markerAnnulusRadiusCmd(int id, const char* lev, CoordFormat format)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      // it may shrink
      update(PIXMAP, m->getAllBBox());

      int cnt = 0;
      double radii[MAXANNULI];
      istrstream str(lev);
      while ((cnt<MAXANNULI) && (str >> radii[cnt]))
	++cnt;
	
      for (int i=0; i<cnt; i++)
	radii[i] = mapLenToRef(radii[i],format);

      ((Annulus*)m)->setRadii(radii,cnt);
      update(PIXMAP, m->getAllBBox());
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void Frame::markerBackCmd()
{
  Marker* m = markers.tail();
  while (m) {
    if (m->isSelected()) {
      Marker* prev = markers.extractPrev(m);
      markers.append(m);
      update(PIXMAP, m->getAllBBox());
      m = prev;
    }
    else
      m=m->previous();
  }
}

void Frame::markerBackCmd(int id)
{
  Marker* m = markers.tail();
  while (m) {
    if (m->getId() == id) {
      markers.extractPrev(m);
      markers.append(m);
      update(PIXMAP, m->getAllBBox());
      return;
    }
    else
      m=m->previous();
  }
}

void Frame::markerBoxAnnulusSizeCmd(int id, const Vector& inner,
				    const Vector& outer, int num,
				    CoordFormat format)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      // it may shrink
      update(PIXMAP, m->getAllBBox());
      Vector s1 = mapLenToRef(inner,format);
      Vector s2 = mapLenToRef(outer,format);
      ((BoxAnnulus*)(m))->setSize(s1,s2,num);
      update(PIXMAP, m->getAllBBox());
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void Frame::markerBoxAnnulusSizeCmd(int id,const char* lev,CoordFormat format)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      // it may shrink
      update(PIXMAP, m->getAllBBox());

      int cnt = 0;
      Vector size[MAXANNULI];
      istrstream str(lev);
      while ((cnt<MAXANNULI) && (str >> size[cnt][0]))
	str >> size[cnt++][1];
	
      // verify proper ratios
      {
	for (int i=0; i<cnt; i++)
	  size[i][1] = size[i][0]*size[cnt-1][1]/size[cnt-1][0];
      }

      // map to ref coord sys
      {
	for (int i=0; i<cnt; i++)
	  size[i] = mapLenToRef(size[i],format);
      }

      ((BoxAnnulus*)m)->setSize(size,cnt);
      update(PIXMAP, m->getAllBBox());
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void Frame::markerBoxAnnulusCreateSizeCmd(int id, const Vector& v)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      printIntegerTcl(((BoxAnnulus*)m)->addSize(v*canvasToRef));
      update(PIXMAP, m->getBBox());
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void Frame::markerBoxAnnulusDeleteSizeCmd(int id, int h)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      if (m->canEdit()) {
	// it may shrink
	update(PIXMAP, m->getAllBBox());
	((BoxAnnulus*)m)->deleteSize(h);
	update(PIXMAP, m->getAllBBox());
      }
      return;
    }
    m=m->next();
  }
}

void Frame::markerCallBackCmd(int id, Marker::Callback cb, const char* p, 
			      const char* a)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      result = m->setCallBack(cb, p, a);
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void Frame::markerCircleRadiusCmd(int id, double r, CoordFormat format)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      // it may shrink
      update(PIXMAP, m->getAllBBox());
      double rr = mapLenToRef(r,format);
      ((Circle*)m)->setRadius(rr);
      update(PIXMAP, m->getAllBBox());
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void Frame::markerColorCmd(const char* clr)
{
  Marker* m=markers.head();
  while (m) {
    if (m->isSelected()) {
      m->setColor(clr);
      update(PIXMAP, m->getAllBBox());
    }
    m=m->next();
  }
}

void Frame::markerColorCmd(int id, const char* clr)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      m->setColor(clr);
      update(PIXMAP, m->getAllBBox());
      return;
    }
    m=m->next();
  }
}

void Frame::markerCommandCmd(MarkerFormat fm, const char* ccmd)
{
  char buf[1024];
  ostrstream ostr(buf,1024);
  ostr << ccmd << endl << ends; // lets terminate it in case

  istrstream istr(buf,1024);
  parseMarker(fm, istr);
}
 
void Frame::markerDeleteCallBackCmd(int id, Marker::Callback cb, const char* p)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      result = m->deleteCallBack(cb, p);
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void Frame::markerDeleteCmd()
{
  Marker* m = markers.head();
  while (m) {
    if (m->isSelected() && m->canDelete()) {
      Marker* next = markers.extractNext(m);
      update(PIXMAP, m->getAllBBox());
      delete m;
      m = next;
    }
    else
      m=m->next();
  }
}

void Frame::markerDeleteCmd(int id)
{
  Marker* m = markers.head();
  while (m) {
    if (m->getId() == id) {
      markers.extractNext(m);
      update(PIXMAP, m->getAllBBox());
      delete m;
      return;
    }
    else
      m=m->next();
  }
}

void Frame::markerDeleteAllCmd()
{
  Marker* m = markers.head();
  while (m) {
    if (m->canDelete()) {
      update(PIXMAP, m->getAllBBox());
      Marker* next = markers.extractNext(m);
      delete m;
      m = next;
    }
    else
      m=m->next();
  }
}

void Frame::markerDeleteLastCmd()
{
  Marker* m=markers.tail();
  if (m) {
    markers.extractNext(m);
    update(PIXMAP, m->getAllBBox());
    delete m;
    return;
  }
}

void Frame::markerEditBeginCmd(int id, int h)
{
  // remember which marker is being edited

  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id && m->canEdit()) {
      editMarker = m;
      editMarker->editBegin(h);
      return;
    }
    m=m->next();
  }

  editMarker = NULL;
}

void Frame::markerEditBeginCmd(const Vector& v, int h)
{
  // remember which marker is being edited

  Marker* m=markers.head();
  while (m) {
    if (m->isSelected() && m->canEdit()) {
      editMarker = m;
      editMarker->editBegin(h);
      return;
    }
    m=m->next();
  }

  editMarker = NULL;
}

void Frame::markerEditMotionCmd(const Vector& v, int h)
{
  if (editMarker) {

    // erase current marker now

    redraw(editMarker->getBBox());
    forceUpdate();

    editMarker->edit(v*canvasToRef, h);
    editMarker->drawXOR();
  }
}

void Frame::markerEditEndCmd()
{
  if (editMarker)
    editMarker->editEnd();
  editMarker = NULL;

  update(PIXMAP);
}

void Frame::markerEllipseRadiusCmd(int id, const Vector& radius, 
				   CoordFormat format)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      // it may shrink
      update(PIXMAP, m->getAllBBox());
      ((Ellipse*)m)->setRadius(mapLenToRef(radius,format));
      update(PIXMAP, m->getAllBBox());
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void Frame::markerEllipseAngleCmd(int id, double angle)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      // it may shrink
      update(PIXMAP, m->getAllBBox());
      ((BaseEllipse*)(m))->setAngle(angle);
      update(PIXMAP, m->getAllBBox());
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void Frame::markerEllipseAnnulusRadiusCmd(int id, const Vector& inner,
					  const Vector& outer,
					  int num,
					  CoordFormat format)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      // it may shrink
      update(PIXMAP, m->getAllBBox());
      Vector r1 = mapLenToRef(inner,format);
      Vector r2 = mapLenToRef(outer,format);
      ((EllipseAnnulus*)(m))->setRadii(r1,r2,num);
      update(PIXMAP, m->getAllBBox());
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void Frame::markerEllipseAnnulusRadiusCmd(int id, const char* lev,
					  CoordFormat format)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      // it may shrink
      update(PIXMAP, m->getAllBBox());

      int cnt = 0;
      Vector radii[MAXANNULI];
      istrstream str(lev);
      while ((cnt<MAXANNULI) && (str >> radii[cnt][0]))
	str >> radii[cnt++][1];
	
      // verify proper ratios
      {
	for (int i=0; i<cnt; i++)
	  radii[i][1] = radii[i][0]*radii[cnt-1][1]/radii[cnt-1][0];
      }

      // map to ref coord sys
      {
	for (int i=0; i<cnt; i++)
	  radii[i] = mapLenToRef(radii[i],format);
      }

      ((EllipseAnnulus*)(m))->setRadii(radii,cnt);
      update(PIXMAP, m->getAllBBox());
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void Frame::markerEllipseAnnulusCreateRadiusCmd(int id, const Vector& v)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      printIntegerTcl(((EllipseAnnulus*)m)->addRadii(v*canvasToRef));
      update(PIXMAP, m->getBBox());
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void Frame::markerEllipseAnnulusDeleteRadiusCmd(int id, int h)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      if (m->canEdit()) {
	// it may shrink
	update(PIXMAP, m->getAllBBox());
	((EllipseAnnulus*)m)->deleteRadii(h);
	update(PIXMAP, m->getAllBBox());
      }
      return;
    }
    m=m->next();
  }
}

void Frame::markerFontCmd(const char* f)
{
  Marker* m=markers.head();
  while (m) {
    if (m->isSelected()) {
      // things can shrink, so do before and after
      update(PIXMAP, m->getAllBBox());
      m->setFont(f);
      update(PIXMAP, m->getAllBBox());
    }
    m=m->next();
  }
}

void Frame::markerFontCmd(int id, const char* f)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      // things can shrink, so do before and after
      update(PIXMAP, m->getAllBBox());
      m->setFont(f);
      update(PIXMAP, m->getAllBBox());
      return;
    }
    m=m->next();
  }
}

void Frame::markerFrontCmd()
{
  Marker* m = markers.head();
  while (m) {
    if (m->isSelected()) {
      Marker* next = markers.extractNext(m);
      markers.insertHead(m);
      update(PIXMAP, m->getAllBBox());
      m = next;
    }
    else
      m=m->next();
  }
}

void Frame::markerFrontCmd(int id)
{
  Marker* m = markers.head();
  while (m) {
    if (m->getId() == id) {
      markers.extractNext(m);
      markers.insertHead(m);
      update(PIXMAP, m->getAllBBox());
      return;
    }
    else
      m=m->next();
  }
}

void Frame::markerLineCmd(int id, const Vector& p1, const Vector& p2, 
			  CoordSystem sys)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      // it may shrink
      update(PIXMAP, m->getAllBBox());
      ((Line*)(m))->setP1(map(p1,sys,REF));
      ((Line*)(m))->setP2(map(p2,sys,REF));
      update(PIXMAP, m->getAllBBox());
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void Frame::markerLineArrowCmd(int id, int p1, int p2)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      ((Line*)(m))->setP1Arrow(p1);
      ((Line*)(m))->setP2Arrow(p2);
      update(PIXMAP, m->getAllBBox());
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void Frame::markerListCmd(MarkerFormat type, CoordSystem sys, 
			  CoordFormat format, unsigned short prop, int which,
			  Delim dl)
{
  char buf[1024];

  char delim[4];
  switch (dl) {
  case SEMICOLON:
    strcpy(delim, "; ");
    break;
  case NL:
    strcpy(delim, "\n");
    break;
  }

  switch (type) {
  case DS9:
    {
      ostrstream str(buf,1024);
      markerListHeader(str, sys, delim);
      str << ends;
      Tcl_AppendResult(interp, buf, NULL);

      Marker* m=markers.head();
      while (m) {
	if (!prop || (m->getProperty(prop) == which)) {
	  ostrstream str(buf,1024);
	  m->list(str, sys, format, delim);
	  str << ends;
	  Tcl_AppendResult(interp, buf, NULL);
	}
	m=m->next();
      }
    }
    break;
  case PROS:
    {
      Marker* m=markers.head();
      while (m) {
	if (!prop || (m->getProperty(prop) == which)) {
	  ostrstream str(buf,1024);
	  m->listPros(str, sys, format, delim);
	  str << ends;
	  Tcl_AppendResult(interp, buf, NULL);
	}
	m=m->next();
      }
    }
    break;
  }
}

void Frame::markerListCmd(MarkerFormat type, CoordSystem sys, 
			  CoordFormat format, Delim dl)
{
  char buf[1024];

  char delim[4];
  switch (dl) {
  case SEMICOLON:
    strcpy(delim, "; ");
    break;
  case NL:
    strcpy(delim, "\n");
    break;
  }

  switch (type) {
  case DS9:
    {
      ostrstream str(buf,1024);
      markerListHeader(str, sys, delim);
      str << ends;
      Tcl_AppendResult(interp, buf, NULL);

      Marker* m=markers.head();
      while (m) {
	ostrstream str(buf,1024);
	m->list(str, sys, format, delim);
	str << ends;
	Tcl_AppendResult(interp, buf, NULL);
	m=m->next();
      }
    }
    return;
  case CIAO:
    {
      ostrstream str(buf,1024);
      markerListCiaoHeader(str, delim);
      str << ends;
      Tcl_AppendResult(interp, buf, NULL);

      Marker* m=markers.head();
      while (m) {
	ostrstream str(buf,1024);
	m->listCiao(str, sys, format, delim);
	str << ends;
	Tcl_AppendResult(interp, buf, NULL);
	m=m->next();
      }
    }
    return;
  case FTOOLS:
    {
      Marker* m=markers.head();
      while (m) {
	ostrstream str(buf,1024);
	m->listFtools(str, sys, format, delim);
	str << ends;
	Tcl_AppendResult(interp, buf, NULL);
	m=m->next();
      }
    }
    return;
  case SAOTNG:
    {
      ostrstream str(buf,1024);
      markerListSAOtngHeader(str, sys, format, delim);
      str << ends;
      Tcl_AppendResult(interp, buf, NULL);

      Marker* m=markers.head();
      while (m) {
	ostrstream str(buf,1024);
	m->listSAOtng(str, sys, format, delim);
	str << ends;
	Tcl_AppendResult(interp, buf, NULL);
	m=m->next();
      }
    }
    return;
  case SAOIMAGE:
    {
      Marker* m=markers.head();
      while (m) {
	ostrstream str(buf,1024);
	m->listSAOimage(str, delim);
	str << ends;
	Tcl_AppendResult(interp, buf, NULL);
	m=m->next();
      }
    }
    return;
  case PROS:
    {
      Marker* m=markers.head();
      while (m) {
	ostrstream str(buf,1024);
	m->listPros(str, sys, format, delim);
	str << ends;
	Tcl_AppendResult(interp, buf, NULL);
	m=m->next();
      }
    }
    return;
  }
}

void Frame::markerListHeader(ostream& str, CoordSystem sys, const char* delim)
{
  str << "# Region file format: DS9 version 3.0" << delim;

  str << "global color=green font=\"helvetica 10 normal\" "
      << "select=1 edit=1 move=1 delete=1 include=1 fixed=0" << delim;
}

void Frame::markerListCiaoHeader(ostream& str, const char* delim)
{
  str << "# Region file format: CIAO version 1.0" << delim;
}

void Frame::markerListSAOtngHeader(ostream& str, CoordSystem sys, 
				   CoordFormat format, const char* delim)
{
  if (currentFits)
    str << "# filename: " << currentFits->getFileName() << delim;

  switch (sys) {
  case PHYSICAL:
    str << "# format: pixels (physical)" << delim;
    break;
  case EQUATORIAL_FK4:
  case EQUATORIAL_FK5:
  case ECLIPTIC:
  case GALACTIC:
  case LINEAR:
  case ICRS:
    str << "# format: ";
    switch (format) {
    case DEGREES:
      str << "degrees (";
      break;
    case SEXAGESIMAL:
      str << "hms (";
      break;
    }

    switch (sys) {
    case Frame::EQUATORIAL_FK4:
      str << "fk4";
      break;
    case Frame::EQUATORIAL_FK5:
      str << "fk5";
      break;
    case Frame::ICRS:
      str << "icrs";
      break;
    case Frame::GALACTIC:
      str << "galactic";
      break;
    case Frame::ECLIPTIC:
      str << "ecliptic";
      break;
    case Frame::LINEAR:
      str << "linear";
      break;
    }

    str << ')' << delim;
    break;
  }
}

void Frame::markerLoadCmd(MarkerFormat fm, const char* fn)
{
  ifstream str(fn);
  if (!str) {
    result = TCL_ERROR;
    return;
  }  

  parseMarker(fm, str);
}

void Frame::markerLoadCmd(MarkerFormat fm, int fd)
{
  ifstream str(fd);
  if (!str) {
    result = TCL_ERROR;
    return;
  }  

  parseMarker(fm, str);
}

void Frame::markerLoadFitsCmd(const char* fn)
{
  // verify that we have an ext specified
  if (fn && (fn[strlen(fn)-1] != ']')) {
    result = TCL_ERROR;
    return;
  }

  FitsFile* mk = new FitsFitsMMap(fn, FitsFile::EXACT);
  if (!mk->isValid() || !mk->isTable()) {
    delete mk;
    result = TCL_ERROR;
  }

  FitsHead* mkh = mk->head();
  if (mkh->naxes() != 2) {
    delete mk;
    result = TCL_ERROR;
  }
  FitsTableHDU* mkhdu = (FitsTableHDU*)mkh->hdu();

  FitsBinColumnA* shape = (FitsBinColumnA*)mkhdu->find("shape");
  FitsColumn* x = mkhdu->find("x");
  FitsColumn* y = mkhdu->find("y");
  FitsColumn* r = mkhdu->find("r");
  FitsColumn* ang = mkhdu->find("rotang");
  
  if (!shape || !x || !y || !r || !ang) {
    delete mk;
    result = TCL_ERROR;
    return;
  }

  char* ptr = (char*)mk->data();
  int rows = mkhdu->rows();
  int rowlen = mkhdu->width();

  char color[] = "green";
  char font[] = "helvetica 10 normal";
  char text[] = "";
  unsigned short props = Marker::SELECT | Marker::EDIT | Marker::MOVE |
    Marker::ROTATE | Marker::DELETE | Marker::INCLUDE;

  for (int i=0; i<rows; i++, ptr+=rowlen) {
    char* s1 = shape->str(ptr);
    char* s2 = toUpper(s1);
    if (!strncmp(s2, "CIRCLE", 6))
      createCircleCmd(Vector(x->value(ptr,0), y->value(ptr,0)), PHYSICAL,
		      r->value(ptr), PIXELS, color, font, text, props);

    else if (!strncmp(s2, "ANNULU", 6))
      createAnnulusCmd(Vector(x->value(ptr,0), y->value(ptr,0)), PHYSICAL,
		       r->value(ptr,0), r->value(ptr,1), 1, PIXELS, 
		       color, font, text, props);

    else if (!strncmp(s2, "BOX   ", 6))
      createBoxCmd(Vector(x->value(ptr,0), y->value(ptr,0)), PHYSICAL,
		   Vector(r->value(ptr,0), r->value(ptr,1)), PIXELS, 
		   0, color, font, text, props);

    else if (!strncmp(s2, "ROTBOX", 6))
      createBoxCmd(Vector(x->value(ptr,0), y->value(ptr,0)), PHYSICAL,
		   Vector(r->value(ptr,0), r->value(ptr,1)), PIXELS, 
		   degToRad(ang->value(ptr)), color, font, text, props);

    else if (!strncmp(s2, "RECTAN", 6)) {
      Vector v1(x->value(ptr,0), x->value(ptr,1));
      Vector v2(y->value(ptr,0), y->value(ptr,1));
      Vector d = v2-v1;
      Vector c = d/2 + v1;

      createBoxCmd(c, PHYSICAL, d, PIXELS, 0, color, font, text, props);
    }

    else if (!strncmp(s2, "ROTREC", 6)) {
      Vector v1(x->value(ptr,0), x->value(ptr,1));
      Vector v2(y->value(ptr,0), y->value(ptr,1));
      Vector d = v2-v1;
      Vector c = d/2 + v1;

      createBoxCmd(c, PHYSICAL, d, PIXELS, degToRad(ang->value(ptr)), 
		   color, font, text, props);
    }

    else if (!strncmp(s2, "ELLIPS", 6))
      createEllipseCmd(Vector(x->value(ptr,0), y->value(ptr,0)), PHYSICAL,
		       Vector(r->value(ptr,0), r->value(ptr,1)), PIXELS, 
		       degToRad(ang->value(ptr)), color, font, text, props);

    else if (!strncmp(s2, "POINT ", 6))
      createBoxCirclePointCmd(Vector(x->value(ptr,0), y->value(ptr,0)),
			      PHYSICAL, color, font, text, props);

    else if (!strncmp(s2, "POLYGO", 6)) {
      List<Vertex>* list = new List<Vertex>();
      Vertex* first = new Vertex(x->value(ptr,0), y->value(ptr,0));
      list->append(first);
      for (int i=1; i<12; i++) {
	Vertex* n = new Vertex(x->value(ptr,i), y->value(ptr,i));
	if (n->vector[0] != first->vector[0] || 
	    n->vector[1] != first->vector[1])
	  list->append(new Vertex(x->value(ptr,i), y->value(ptr,i)));
	else
	  break;
      }

      createPolygonCmd(list, PHYSICAL, color, font, text, props);
    }

    delete s1;
    delete s2;
  }

  delete mk;
}

void Frame::markerMoveCmd(const Vector& v)
{
  Marker* m=markers.head();
  while (m) {
    if (m->isSelected() && m->canMove()) {
      Vector c = m->getCenter() * refToCanvas;
      update(PIXMAP, m->getAllBBox());
      m->moveTo((c + v) * canvasToRef);
      update(PIXMAP, m->getAllBBox());
    }
    m=m->next();
  }
}

void Frame::markerMoveCmd(int id, const Vector& v)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() && m->canMove()) {
      Vector c = m->getCenter() * refToCanvas;
      update(PIXMAP, m->getAllBBox());
      m->moveTo((c + v) * canvasToRef);
      update(PIXMAP, m->getAllBBox());
      return;
    }
    m=m->next();
  }
}

void Frame::markerMoveBeginCmd(const Vector& v)
{
  markerBegin = v * canvasToRef;
}

void Frame::markerMoveBeginCmd(int id, const Vector& v)
{
  markerBegin = v * canvasToRef;
}

void Frame::markerMoveMotionCmd(const Vector& v)
{
  // first, accumulate erase markers

  Marker* m=markers.head();
  if (m) {
    while (m) {
      if (m->isSelected() && m->canMove())
	redraw(m->getBBox());
      m=m->next();
    }

    // and erase now

    forceUpdate();

    // ok, now draw selected markers in new location

    Vector markerEnd = v * canvasToRef;
    Vector diff = markerEnd - markerBegin;
    markerBegin = markerEnd;

    m=markers.head();
    while (m) {
      if (m->isSelected() && m->canMove()) {
	m->move(diff);
	m->drawXOR();
      }
      m=m->next();
    }
  }
}

void Frame::markerMoveMotionCmd(int id, const Vector& v)
{
  // first, accumulate erase markers

  Marker* m=markers.head(); 
  if (m) {
    while (m) {
      if (m->getId() && m->canMove()) {
	redraw(m->getBBox());
	break;
      }
      m=m->next();
    }

    if (!m)
      return; // can't find it

    Marker *ptr = m;

    // and erase now

    forceUpdate();

    // ok, now draw selected markers in new location

    Vector markerEnd = v * canvasToRef;
    Vector diff = markerEnd - markerBegin;
    markerBegin = markerEnd;

    ptr->move(diff);
    ptr->drawXOR();
  }
}

void Frame::markerMoveEndCmd()
{
  // update widget since we don't know where the selected markers came from
  update(PIXMAP);
}

void Frame::markerMoveEndCmd(int id)
{
  // update widget since we don't know where the selected markers came from
  update(PIXMAP);
}

void Frame::markerMoveToCmd(const Vector& v, CoordSystem sys)
{
  Marker* m=markers.head();
  while (m) {
    if (m->isSelected() && m->canMove()) {
      update(PIXMAP, m->getAllBBox());
      m->moveTo(map(v,sys,REF));
      update(PIXMAP, m->getAllBBox());
    }
    m=m->next();
  }
}

void Frame::markerMoveToCmd(int id, const Vector& v, CoordSystem sys)
{
  Marker* m=markers.head();
  while (m) {
    if ((m->getId() == id) && m->canMove()) {
      update(PIXMAP, m->getAllBBox());
      m->moveTo(map(v,sys,REF));
      update(PIXMAP, m->getAllBBox());
      return;
    }
    m=m->next();
  }
}

void Frame::markerPolygonAngleCmd(int id, double angle)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      // it may shrink
      update(PIXMAP, m->getAllBBox());
      ((Polygon*)(m))->setAngle(angle);
      update(PIXMAP, m->getAllBBox());
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void Frame::markerPolygonCreateVertexCmd(int id, int seg, const Vector& v)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      ((Polygon*)(m))->createVertex(seg, v * canvasToRef);
      update(PIXMAP, m->getBBox());
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void Frame::markerPolygonDeleteVertexCmd(int id, int h)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      if (m->canEdit()) {
	// it may shrink
	update(PIXMAP, m->getAllBBox());
	((Polygon*)(m))->deleteVertex(h);
	update(PIXMAP, m->getAllBBox());
      }
      return;
    }
    m=m->next();
  }
}

void Frame::markerPropertyCmd(unsigned short prop, int value)
{
  Marker* m=markers.head();
  while (m) {
    if (m->isSelected()) {
      if (prop == Marker::FIXED || prop == Marker::INCLUDE) {
	// marker will change bbox, so get before and after
	update(PIXMAP, m->getAllBBox());
	m->setProperty(prop, value);
	update(PIXMAP, m->getAllBBox());
      }
      else
	m->setProperty(prop, value);
    }
    m=m->next();
  }
}

void Frame::markerPropertyCmd(int id, unsigned short prop, int value)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      if (prop == Marker::FIXED || prop == Marker::INCLUDE) {
	// marker will change bbox, so get before and after
	update(PIXMAP, m->getAllBBox());
	m->setProperty(prop, value);
	update(PIXMAP, m->getAllBBox());
      }
      else
	m->setProperty(prop, value);

      return;
    }
    m=m->next();
  }
}

void Frame::markerBoxSizeCmd(int id, const Vector& size, CoordFormat format)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      // it may shrink
      update(PIXMAP, m->getAllBBox());
      ((Box*)(m))->setSize(mapLenToRef(size,format));
      update(PIXMAP, m->getAllBBox());
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void Frame::markerBoxAngleCmd(int id, double angle)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      // it may shrink
      update(PIXMAP, m->getAllBBox());
      ((Box*)(m))->setAngle(angle);
      update(PIXMAP, m->getAllBBox());
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void Frame::markerRotateBeginCmd(int id)
{
  // remember which marker is being edited

  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id && m->canRotate()) {
      rotateMarker = m;
      return;
    }
    m=m->next();
  }

  rotateMarker = NULL;
}

void Frame::markerRotateBeginCmd(const Vector& v)
{
  // remember which marker is being edited

  Marker* m=markers.head();
  while (m) {
    if (m->isSelected() && m->canRotate()) {
      rotateMarker = m;
      return;
    }
    m=m->next();
  }

  rotateMarker = NULL;
}

void Frame::markerRotateMotionCmd(const Vector& v, int h)
{
  if (rotateMarker) {

    // erase current marker now

    redraw(rotateMarker->getBBox());
    forceUpdate();

    rotateMarker->rotate(v*canvasToRef, h);
    rotateMarker->drawXOR();
  }
}

void Frame::markerRotateEndCmd()
{
  rotateMarker = NULL;
  update(PIXMAP);
}

void Frame::markerRulerPointCmd(int id, const Vector& p1, const Vector& p2,
				CoordSystem sys)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      // it may shrink
      update(PIXMAP, m->getAllBBox());
      ((Ruler*)(m))->setP1(map(p1,sys,REF));
      ((Ruler*)(m))->setP2(map(p2,sys,REF));
      update(PIXMAP, m->getAllBBox());
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void Frame::markerRulerSystemCmd(int id, CoordSystem sys, CoordFormat format)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      // it may shrink
      update(PIXMAP, m->getAllBBox());
      ((Ruler*)(m))->setSystem(sys, format);
      update(PIXMAP, m->getAllBBox());
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void Frame::markerSaveCmd(const char* fileName, MarkerFormat type, 
			  CoordSystem sys, CoordFormat format)
{
  char delim[] = "\n";

  ofstream fn(fileName);
  if (fn) {
    switch (type) {
    case DS9:
      {
	markerListHeader(fn, sys, delim);
	Marker* m=markers.head();
	while (m) {
	  m->list(fn, sys, format, delim);
	  m=m->next();
	}
      }
      break;
    case CIAO:
      {
	markerListCiaoHeader(fn, delim);
	Marker* m=markers.head();
	while (m) {
	  m->listCiao(fn, sys, format, delim);
	  m=m->next();
	}
      }
      return;
    case FTOOLS:
      {
	Marker* m=markers.head();
	while (m) {
	  m->listFtools(fn, sys, format, delim);
	  m=m->next();
	}
      }
      return;
    case SAOTNG:
      {
	markerListSAOtngHeader(fn, sys, format, delim);
	Marker* m=markers.head();
	while (m) {
	  m->listSAOtng(fn, sys, format, delim);
	  m=m->next();
	}
      }
      return;
    case SAOIMAGE:
      {
	Marker* m=markers.head();
	while (m) {
	  m->listSAOimage(fn, delim);
	  m=m->next();
	}
      }
      return;
    case PROS:
      {
	Marker* m=markers.head();
	while (m) {
	  m->listPros(fn, sys, format, delim);
	  m=m->next();
	}
      }
      return;
    }
  }
  else {
    Tcl_AppendResult(interp, "Unable to open file ", fileName, NULL);
    result = TCL_ERROR;
  }
}

void Frame::markerSelectCmd(int id)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      m->select();
      update(PIXMAP, m->getBBox());
      return;
    }
    m=m->next();
  }
}

void Frame::markerSelectCmd(const Vector& v)
{
  // first, check to see if we clicked on an already selected marker

  Marker* m=markers.head();
  while (m) {
    if (m->isIn(v) && m->isSelected()) {
      Tcl_AppendResult(interp, "1", NULL);
      return;
    }
    m=m->next();
  }

  // ok, now select the first found, and unselect the rest

  int found = 0;

  m=markers.head();
  while (m) {
    if (m->isIn(v) && m->canSelect() && !found) {
      if (!m->isSelected()) {
	m->select();
	update(PIXMAP, m->getBBox());
      }
      found = 1;
    }
    else {
      if (m->isSelected()) {
	m->unselect();
	update(PIXMAP, m->getBBox());
      }
    }
    m=m->next();
  } 

  if (found)
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
}

void Frame::markerSelectAllCmd()
{
  Marker* m=markers.head();
  while (m) {
    if (m->canSelect()) {
      m->select();
      update(PIXMAP, m->getBBox());
    }
    m=m->next();
  }
}

void Frame::markerSelectToggleCmd(const Vector& v)
{
  // toggle the select of the first found

  Marker* m=markers.head();
  while (m) {
    if (m->isIn(v) && m->canSelect()) {
      m->toggleSelect();
      update(PIXMAP, m->getBBox());
      Tcl_AppendResult(interp, "1", NULL);
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "0", NULL);
}

void Frame::markerTextCmd(int id, const char* text)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      // things can shrink, so do before and after
      update(PIXMAP,m->getAllBBox());
      m->setText(text);
      update(PIXMAP,m->getAllBBox());
      return;
    }
    m=m->next();
  }
}

void Frame::markerUnselectCmd(int id)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      m->unselect();
      update(PIXMAP, m->getBBox());
      return;
    }
    m=m->next();
  }
}

void Frame::markerUnselectAllCmd()
{
  Marker* m=markers.head();
  while (m) {
    m->unselect();
    update(PIXMAP, m->getBBox());
    m=m->next();
  } 
}

void Frame::regionSelectBeginCmd(const Vector& v)
{
  regionBegin = v;
  regionEnd = v;
}

void Frame::regionSelectMotionCmd(const Vector& v)
{
  if (visible) {

    // erase 

    redrawNow((BBox(regionBegin, regionEnd)).expand(2));

    // and draw to window

    regionEnd = v;
    BBox cc = BBox(regionBegin, regionEnd) * canvasToWindow;
    Vector size = cc.size();

    XDrawRectangle(display, Tk_WindowId(tkwin), markerGCXOR,
		   (int)cc.ll[0], (int)cc.ll[1], (int)size[0], (int)size[1]);
  }
}

void Frame::regionSelectEndCmd()
{
  BBox bb(regionBegin, regionEnd);

  Marker* m=markers.head();
  while (m) {
    if (bb.isIn(m->getBBox())==4 && m->canSelect())
      m->select();
    else
      m->unselect();
    m=m->next();
  } 

  update(PIXMAP, bb.expand(2));
}

void Frame::regionSelectShiftEndCmd()
{
  BBox bb(regionBegin, regionEnd);

  Marker* m=markers.head();
  while (m) {
    if (bb.isIn(m->getBBox())==4)
      m->select();
    m=m->next();
  }

  update(PIXMAP, bb.expand(2));
}

// Marker Support

void Frame::parseMarker(MarkerFormat fm, istream& str)
{
  switch (fm) {
  case DS9:
  case CIAO:
    mklexx = new mkFlexLexer(&str);
    mkmode = fm;
    mkparse(this);
    delete mklexx;
    mklexx = NULL;
    break;
  case FTOOLS:
  case SAOTNG:
  case SAOIMAGE:
  case PROS:
    othlexx = new othFlexLexer(&str);
    othparse(this);
    delete othlexx;
    othlexx = NULL;
    break;
  }
}

void Frame::psMarkers(int mode)
{
  // render from back to front
  // bbox is in canvas coords

  const BBox bb = BBox(0, 0, options->width-1, options->height-1) * 
    widgetToCanvas;

  Marker* m=markers.tail();
  while (m) {
    if (m->isVisible(bb))
      m->ps(mode);
    m=m->previous();
  }
}

void Frame::renderMarkers(const BBox& bb)
{
  // render from back to front
  // bbox is in canvas coords

  Marker* m=markers.tail();
  while (m) {
    if (m->isVisible(bb))
      m->draw();
    m=m->previous();
  }
}

void Frame::renderMagnifierMarkers(const Matrix& matrix, const BBox& bb)
{
  // render from back to front
  // bbox is in canvas coords

  Marker* m=markers.tail();
  while (m) {
    if (m->isVisible(bb))
      m->drawMagnifier(matrix);
    m=m->previous();
  }
}

void Frame::unselectMarkers()
{
  Marker* m=markers.head();
  while (m) {
    m->unselect();
    m=m->next();
  }
}

void Frame::updateMarkers()
{
  Marker* m=markers.head();
  while (m) {
    m->updateBBox();
    m=m->next();
  }
}

void Frame::updateMarkerCoords(Matrix mat)
{
  Marker* m=markers.head();
  while (m) {
    m->updateCoords(mat);
    m=m->next();
  }
}
