#include "line.h"
#include "frame.h"


Line::Line(Frame* p, const Vector& ctr, const Vector& ptr, int a1, int a2,
	   const char* clr, const char* f, const char* t, unsigned short prop) 
  : Marker(p, ctr, clr, f, t, prop)
{
  p1Arrow = a1;
  p2Arrow = a2;

  p2 = ptr;
  strcpy(type,"line");
  handle = new Vector[2];
  numHandle = 2;

  updateBBox();
}

double Line::getDistance(int coordFormat)
{
  return parent->mapLenFromRef(p2, center, (Frame::CoordFormat)coordFormat);
}

void Line::updateBBox()
{
  // bound marker

  bbox = BBox(center * parent->refToCanvas, p2 * parent->refToCanvas);

  // generate handles in canvas coords

  handle[0] = center * parent->refToCanvas;
  handle[1] = p2 * parent->refToCanvas;

  // make room for handles/arrows

  bbox.expand(5);

  // calculate overall bbox

  calcAllBBox();
}

void Line::updateCoords(Matrix& m)
{
  p2*=m;
  Marker::updateCoords(m);
}

void Line::draw()
{
  Marker::draw();
  includeExclude();
  render(parent->pixmap, parent->refToWidget, gc);
}

void Line::drawXOR()
{
  render(Tk_WindowId(parent->tkwin), parent->refToWindow, gcxor);
}

void Line::drawMagnifier(const Matrix& m)
{
  XSetForeground(display, gc, color);
  includeExclude();
  render(parent->magnifierPixmap, m, gc);
}

void Line::move(const Vector& v)
{
  center+=v;
  p2+=v;
  updateBBox();

  doCallBacks(&moveCB);
}

void Line::moveTo(const Vector& v)
{
  Vector diff = v - center;
  center=v;
  p2+=diff;
  updateBBox();

  doCallBacks(&moveCB);
}

void Line::edit(const Vector& v, int h)
{
  switch (h) {
  case 1:
    center = v;
    break;
  case 2:
    p2 = v;
    break;
  }
  updateBBox();
  doCallBacks(&editCB);
}

void Line::ps(int mode)
{
  Vector a = center * parent->refToCanvas;
  Vector b = p2 * parent->refToCanvas;

  Marker::ps(mode);
  psIncludeExclude();

  char buf[256];
  ostrstream str(buf,256);
  str << "newpath " 
      << a.TkCanvasPs(parent->canvas) << "moveto"
      << b.TkCanvasPs(parent->canvas) << "lineto"
      << " stroke" << endl << ends;
  Tcl_AppendResult(parent->interp, buf, NULL);

  if (p1Arrow)
    psArrow(a, (b-a).normalize());

  if (p2Arrow)
    psArrow(b, (a-b).normalize());
}

int Line::isIn(const Vector& v)
{
  // do this in canvas coords, not ref coords

  Vector l1 = center*parent->refToCanvas;
  Vector l2 = p2*parent->refToCanvas;
  double a = (l2-l1).angle();

  Matrix m = Translate(-l1) * Rotate(a);
  Vector end = l2*m;
  Vector vv = v*m;
  return (vv[0]>0 && vv[0]<end[0] && vv[1]>-3 && vv[1]<3);
}

void Line::setP1(const Vector& v)
{
  center = v;
  updateBBox();

  doCallBacks(&editCB);
}

void Line::setP2(const Vector& v)
{
  p2 = v;
  updateBBox();

  doCallBacks(&editCB);
}

void Line::setP1Arrow(int w)
{
  p1Arrow = w ? 1 : 0;
  updateBBox();
}

void Line::setP2Arrow(int w)
{
  p2Arrow = w ? 1 : 0;
  updateBBox();
}

// Private

void Line::render(Drawable drawable, const Matrix& matrix, GC ggc)
{
  Vector a = (center * matrix).round();
  Vector b = (p2 * matrix).round();
  XDrawLine(display, drawable, ggc, (int)a[0], (int)a[1], 
	    (int)b[0], (int)b[1]);

  if (p1Arrow)
    renderArrow(drawable, a, (b-a).normalize(), ggc);

  if (p2Arrow)
    renderArrow(drawable, b, (a-b).normalize(), ggc);
}

// list

void Line::list(ostream& str, int sys, int format, const char* delim)
{
  listCoordSystem(str, sys);
  str << ';';

  if (!(properties&INCLUDE))
    str << '-';

  switch ((Frame::CoordSystem)sys) {
  case Frame::IMAGE:
  case Frame::PHYSICAL:
  case Frame::DETECTOR:
  case Frame::AMPLIFIER:
    {
      Vector v1 = parent->map(center,Frame::REF,(Frame::CoordSystem)sys);
      Vector v2 = parent->map(p2,Frame::REF,(Frame::CoordSystem)sys);
      str << type << '(' << setprecision(8) << v1[0] << ',' << v1[1] << ','
	  << v2[0] << ',' << v2[1] << ')';
    }
    break;

  case Frame::EQUATORIAL_FK4:
  case Frame::EQUATORIAL_FK5:
  case Frame::ECLIPTIC:
  case Frame::GALACTIC:
  case Frame::LINEAR:
  case Frame::ICRS:
    switch ((Frame::CoordFormat)format) {
    case Frame::DEGREES:
      {
	Vector v1 = parent->map(center,Frame::REF,(Frame::CoordSystem)sys);
	Vector v2 = parent->map(p2,Frame::REF,(Frame::CoordSystem)sys);
	str << type << '(' << setprecision(8) << v1[0] << ',' << v1[1] << ','
	    << v2[0] << ',' << v2[1] << ')';
      }
      break;
    case Frame::SEXAGESIMAL:
      {
	char buf[64];
	char ra1[16], ra2[16];
	char dec1[16], dec2[16];
	{
	  parent->mapToWCSStr(center, Frame::REF, (Frame::CoordSystem)sys,
			      (Frame::CoordFormat)format, buf, 64);
	  istrstream wcs(buf,64);
	  wcs >> ra1 >> dec1;
	}
	{
	  parent->mapToWCSStr(p2, Frame::REF, (Frame::CoordSystem)sys,
			      (Frame::CoordFormat)format, buf, 64);
	  istrstream wcs(buf,64);
	  wcs >> ra2 >> dec2;
	}
	str << type << '(' << ra1 << ',' << dec1 << ',' 
	    << ra2 << ',' << dec2 << ')';
      }
      break;
    }
    break;
  }

  listProperties(str);
  str << delim;
}

void Line::listProperties(ostream& str)
{
  str << " # line=" << p1Arrow << ' ' << p2Arrow;
  listProps(str);
}

void Line::listFtools(ostream& str, int sys, int format, const char* delim)
{
  if (!(properties&INCLUDE))
    str << '-';

  Vector v1 = parent->map(center,Frame::REF,(Frame::CoordSystem)sys);
  Vector v2 = parent->map(p2,Frame::REF,(Frame::CoordSystem)sys);
  str << type << '(' << setprecision(8) << v1[0] << ',' << v1[1] << ','
      << v2[0] << ',' << v2[1] << ')';

  str << delim;
}

void Line::listSAOtng(ostream& str, int sys, int format, const char* delim)
{
  if (text[0])
    str << '#' << text << delim;

  if (properties&INCLUDE)
    str << '+';
  else
    str << '-';

  switch ((Frame::CoordSystem)sys) {
  case Frame::IMAGE:
  case Frame::DETECTOR:
  case Frame::AMPLIFIER:
    {
      Vector v1 = parent->map(center,Frame::REF,Frame::IMAGE);
      Vector v2 = parent->map(p2,Frame::REF,Frame::IMAGE);
      str << type << '(' << setprecision(8) << v1[0] << ',' << v1[1] << ','
	  << v2[0] << ',' << v2[1] << ')';
    }
    break;
  case Frame::PHYSICAL:
    {
      Vector v1 = parent->map(center,Frame::REF,Frame::PHYSICAL);
      Vector v2 = parent->map(p2,Frame::REF,Frame::PHYSICAL);
      str << type << '(' << setprecision(8) << v1[0] << ',' << v1[1] << ','
	  << v2[0] << ',' << v2[1] << ')';
    }
    break;

  case Frame::EQUATORIAL_FK4:
  case Frame::EQUATORIAL_FK5:
  case Frame::ECLIPTIC:
  case Frame::GALACTIC:
  case Frame::LINEAR:
  case Frame::ICRS:
    switch ((Frame::CoordFormat)format) {
    case Frame::DEGREES:
      {
	Vector v1 = parent->map(center,Frame::REF,(Frame::CoordSystem)sys);
	Vector v2 = parent->map(p2,Frame::REF,(Frame::CoordSystem)sys);
	str << type << '(' << setprecision(8) << v1[0] << ',' << v1[1] << ','
	    << v2[0] << ',' << v2[1] << ')';
      }
      break;
    case Frame::SEXAGESIMAL:
      {
	char buf[64];
	char ra1[16], ra2[16];
	char dec1[16], dec2[16];
	{
	  parent->mapToWCSStr(center, Frame::REF, (Frame::CoordSystem)sys,
			      (Frame::CoordFormat)format, buf, 64);
	  istrstream wcs(buf,64);
	  wcs >> ra1 >> dec1;
	}
	{
	  parent->mapToWCSStr(p2, Frame::REF, (Frame::CoordSystem)sys,
			      (Frame::CoordFormat)format, buf, 64);
	  istrstream wcs(buf,64);
	  wcs >> ra2 >> dec2;
	}
	str << type << '(' << ra1 << ',' << dec1 << ',' 
	    << ra2 << ',' << dec2 << ')';
      }
      break;
    }
    break;
  }

  str << " # " << colorName << delim;
}
